/* gpx-import/src/gpx.c
 *
 * GPX load and memory management
 *
 * Copyright Daniel Silverstone <dsilvers@digital-scurf.org>
 *
 * Written for the OpenStreetMap project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of the License
 * only.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <expat.h>
#include <stdlib.h>
#include <alloca.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdbool.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <time.h>

#include "gpx.h"

#define GPX_BUFLEN 4096
#define ERR_BUFFER_SIZE 1024

#define PARSE_ERROR(ctx,F...)                           \
  do {                                                  \
    if (ctx->err != NULL && *(ctx->err) == NULL)        \
      gpx_record_error(ctx, F);                         \
  } while (0)

void
gpx_free(GPX *gpx)
{
  GPXTrackPoint *pt = gpx->points, *ptnext;
  while (pt != NULL) {
    ptnext = pt->next;
    free(pt->timestamp);
    free(pt);
    pt = ptnext;
  }
  free(gpx);
}

typedef enum {
  UNKNOWN,
  TRACKPOINT,
  ELEVATION,
  TIMESTAMP,
} GPXParseState;

static const char *
gpx_state_name(GPXParseState p)
{
  switch (p) {
  case UNKNOWN:
    return "UNKNOWN";
  case TRACKPOINT:
    return "TRACKPOINT";
  case ELEVATION:
    return "ELEVATION";
  case TIMESTAMP:
    return "TIMESTAMP";
  }
  return "INVALID";
}

typedef struct {
  XML_Parser p;
  GPX *gpx;
  GPXTrackPoint *point;
  GPXTrackPoint *lastpoint;
  uint32_t curseg;
  char *accumulator;
  int accumulator_size;
  GPXParseState state;
  bool got_lat, got_long, got_ele, got_time;
  char **err;
} GPXParseContext;

static void
gpx_record_error(GPXParseContext *ctx, char *fmt, ...)
{
  va_list va;
  int sz;
  char msg[ERR_BUFFER_SIZE];
  
  va_start(va, fmt);
  sz = vsnprintf(msg, ERR_BUFFER_SIZE, fmt, va);
  va_end(va);
  
  *(ctx->err) = calloc(2, ERR_BUFFER_SIZE);
  snprintf(*(ctx->err), 2 * ERR_BUFFER_SIZE, "%s\n  XML parser at line %ld column %ld",
           msg,
           XML_GetCurrentLineNumber(ctx->p),
           XML_GetCurrentColumnNumber(ctx->p));
}

static GPXCoord
gpx_parse_coord(const XML_Char *str)
{
  GPXCoord ret = 0;
  bool is_neg = false;
  bool is_afterdot = false;
  int afterdot = 0;
  
  if (*str == '-') {
    is_neg = true;
    str++;
  }
  while (*str != '\0') {
    if (*str == '.')
      is_afterdot = true;
    if (*str != '.') {
      ret *= 10;
      ret += (*str - '0');
      if (is_afterdot == true)
        afterdot++;
    }
    str++;
  }
  
  while (afterdot < 9) {
    afterdot++;
    ret *= 10;
  }
  
  return (is_neg ? -ret : ret);
}

static void
gpx_clear_accumulator(GPXParseContext *ctx)
{
  if (ctx->accumulator != NULL)
    free(ctx->accumulator);
  ctx->accumulator = NULL;
  ctx->accumulator_size = 0;
}

#define REQUIRE_STATE(S)                                                \
  if (ctx->state != (S)) {                                              \
    PARSE_ERROR(ctx, "Expected state %s, got state %s", gpx_state_name(S), gpx_state_name(ctx->state)); \
    XML_StopParser(ctx->p, 0);                                          \
    return;                                                             \
  }

static void
gpx_handle_start_element(void *_ctx, const XML_Char *name, const XML_Char **atts)
{
  GPXParseContext *ctx = (GPXParseContext *)(_ctx);
  
  if (strcmp(name, "trkpt") == 0) {
    REQUIRE_STATE(UNKNOWN);
    ctx->state = TRACKPOINT;
    ctx->point = calloc(1, sizeof(GPXTrackPoint));
    ctx->point->segment = ctx->curseg;
    ctx->got_lat = ctx->got_long = ctx->got_ele = ctx->got_time = false;
    while (*atts != NULL) {
      if (strcmp(*atts, "lat") == 0) {
        atts++;
        ctx->point->latitude = gpx_parse_coord(*atts++);
        ctx->got_lat = true;
      } else if (strcmp(*atts, "lon") == 0) {
        atts++;
        ctx->point->longitude = gpx_parse_coord(*atts++);
        ctx->got_long = true;
      } else {
        atts += 2; /* skip tag + value */
      }
    }
  } else if (strcmp(name, "ele") == 0) {
    REQUIRE_STATE(TRACKPOINT);
    ctx->state = ELEVATION;
    gpx_clear_accumulator(ctx);
  } else if (strcmp(name, "time") == 0) {
    if (ctx->state == TRACKPOINT) {
      ctx->state = TIMESTAMP;
      gpx_clear_accumulator(ctx);
    }
  }
}

#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif

static void
gpx_update_minmax(GPX *gpx, GPXTrackPoint *point)
{
  gpx->minlatitude = MIN(gpx->minlatitude, point->latitude);
  gpx->minlongitude = MIN(gpx->minlongitude, point->longitude);
  gpx->maxlatitude = MAX(gpx->maxlatitude, point->latitude);
  gpx->maxlongitude = MAX(gpx->maxlongitude, point->longitude);
}

static void
gpx_handle_end_element(void *_ctx, const XML_Char *name)
{
  GPXParseContext *ctx = (GPXParseContext *)(_ctx);
  
  if (strcmp(name, "trkpt") == 0) {
    REQUIRE_STATE(TRACKPOINT);
    /* Remove the || true if elevation is mandatory. */
    if ((ctx->got_lat && ctx->got_long && (ctx->got_ele || true) && ctx->got_time) &&
        (ctx->point != NULL) && (ctx->point->longitude >= -180000000000) &&
        (ctx->point->longitude <= 180000000000) &&
        (ctx->point->latitude >= -90000000000) &&
        (ctx->point->latitude <= 90000000000)) {
      gpx_update_minmax(ctx->gpx, ctx->point);
      if (ctx->lastpoint == NULL) {
        ctx->gpx->firstlatitude = ctx->point->latitude;
        ctx->gpx->firstlongitude = ctx->point->longitude;
        ctx->lastpoint = ctx->point;
        ctx->gpx->points = ctx->point;
        ctx->point = NULL;
      } else {
        ctx->lastpoint->next = ctx->point;
        ctx->lastpoint = ctx->point;
        ctx->point = NULL;
      }
      ctx->gpx->goodpoints++;
    } else {
      ctx->gpx->badpoints++;
      if (ctx->point->timestamp != NULL)
        free(ctx->point->timestamp);
      free(ctx->point);
      ctx->point = NULL;
    }
    ctx->state = UNKNOWN;
  } else if (strcmp(name, "ele") == 0) {
    REQUIRE_STATE(ELEVATION);
    ctx->point->elevation = strtof(ctx->accumulator, NULL);
    ctx->state = TRACKPOINT;
    ctx->got_ele = true;
  } else if (strcmp(name, "time") == 0) {
    char *pnull = NULL;
    struct tm ignored;
    if (ctx->state == UNKNOWN)
      return;
    REQUIRE_STATE(TIMESTAMP);
    pnull = strptime(ctx->accumulator ? ctx->accumulator : "",
                     "%Y-%m-%dT%H:%M:%SZ",
                     &ignored);
    if (pnull != NULL && *pnull == '\0') {
      ctx->point->timestamp = ctx->accumulator;
      ctx->accumulator = NULL;
      ctx->got_time = true;
    }
    ctx->state = TRACKPOINT;
  } else if (strcmp(name, "trkseg") == 0) {
    REQUIRE_STATE(UNKNOWN);
    ctx->curseg++;
  }
}

void
gpx_handle_string_data(void *_ctx, const XML_Char *str, int len)
{
  GPXParseContext *ctx = (GPXParseContext *)(_ctx);
  
  if ((ctx->state != ELEVATION) &&
      (ctx->state != TIMESTAMP))
    /* We only accumulate during elevation or timestamp */
    return;
  
  if (ctx->accumulator == NULL) {
    ctx->accumulator = malloc(len + 1);
    memcpy(ctx->accumulator, str, len);
    ctx->accumulator_size = len + 1;
  } else {
    ctx->accumulator = realloc(ctx->accumulator, ctx->accumulator_size + len);
    memcpy(ctx->accumulator + ctx->accumulator_size - 1, str, len);
    ctx->accumulator_size += len;
  }
  
  ctx->accumulator[ctx->accumulator_size - 1] = '\0';
}

GPX*
gpx_parse_file(const char *gpxfile, char **err)
{
  GPXParseContext ctx;
  XML_Parser p;
  char *buffer = alloca(GPX_BUFLEN);
  ssize_t bufread;
  int fd;
  
  ctx.err = err;
  ctx.p = p = XML_ParserCreate(NULL);
  
  if (p == NULL)
    return NULL;
  
  XML_SetElementHandler(p, gpx_handle_start_element, gpx_handle_end_element);
  XML_SetDefaultHandler(p, gpx_handle_string_data);
  XML_SetUserData(p, &ctx);
  
  fd = open(gpxfile, O_RDONLY);
  if (fd == -1) {
    PARSE_ERROR((&ctx), "Error opening file %s: %s", gpxfile, strerror(errno));
    XML_ParserFree(p);
    return NULL;
  }
  
  ctx.gpx = calloc(1, sizeof(GPX));
  ctx.gpx->minlatitude = 91000000000;
  ctx.gpx->minlongitude = 181000000000;
  ctx.gpx->maxlatitude = -91000000000;
  ctx.gpx->maxlongitude = -181000000000;
  ctx.point = NULL;
  ctx.lastpoint = NULL;
  ctx.curseg = 0;
  ctx.accumulator = NULL;
  ctx.state = UNKNOWN;
  ctx.accumulator_size = 0;
  
  while ((bufread = read(fd, buffer, GPX_BUFLEN)) > 0) {
    if (XML_Parse(p, buffer, bufread, XML_FALSE) != XML_STATUS_OK) {
      gpx_free(ctx.gpx);
      if (ctx.point != NULL) {
        if (ctx.point->timestamp)
          free(ctx.point->timestamp);
        free(ctx.point);
      }
      if (ctx.accumulator != NULL)
        free(ctx.accumulator);
      XML_ParserFree(p);
      return NULL;
    }
  }
  
  XML_ParserFree(p);
  if (ctx.point)
    free(ctx.point);
  if (ctx.accumulator != NULL)
    free(ctx.accumulator);
  
  return ctx.gpx;
  
}

void
gpx_print(GPX *gpx)
{
  GPXTrackPoint *pt;
  int pointnr;
  
  printf("minlat=%ld maxlat=%ld minlon=%ld maxlon=%ld\n",
         gpx->minlatitude, gpx->maxlatitude,
         gpx->minlongitude, gpx->maxlongitude);
  
  printf("goodpoints=%d badpoints=%d\n",
         gpx->goodpoints,
         gpx->badpoints);
  return;
  for (pt = gpx->points, pointnr = 1; pt != NULL; pt = pt->next, pointnr++)
    printf("%4d: lat=%ld lon=%ld ele=%f time=%s seg=%d\n",
           pointnr, pt->latitude, pt->longitude, pt->elevation,
           pt->timestamp, pt->segment);
  
}
